from __future__ import unicode_literals
from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect, render, get_object_or_404
from django.http import HttpResponse
from django.template import loader
from django.views import generic
from .models import *
from .forms import *
from .auth import * # Own wrapper for django logging in/out
from django.forms import modelformset_factory
from django.shortcuts import render
from django.db.models import Q
import collections
import datetime
from django.db.models import Count
from django.db.models import Case
from django.db.models import When

import csv


# See https://docs.djangoproject.com/en/1.10/topics/http/views/#customizing-error-views
# They should work ONLY in production! (DEBUG=False)
handler404 = 'web.views.e404_page_not_found'
handler500 = 'web.views.e500_error'
handler403 = 'web.views.e403_permission_denied'
handler400 = 'web.views.e400_bad_request'


def index(request):
	if request.user.is_authenticated():
		return redirect(appointments)
	return redirect(login)


def e404_page_not_found(request, context = None):
	return render(request, "errors/404.html", context, status=404)

def e500_error(request, context = None):
	return render(request, "errors/500.html", context, status=500)

def e403_permission_denied(request, context = None):
	return render(request, "errors/403.html", context, status=403)

def e400_bad_request(request, context = None):
	return render(request, "errors/400.html", context, status=400)


def login(request):
	context = {
		'state' : 'initial'
	}
	if request.GET and request.GET.get('error'):
		context['state'] = request.GET.get('error')

	if request.method == "POST" and request.POST:
		state, message = do_login(request)
		if state == True:
			return redirect(appointments)
		else:
			return redirect('/login?error=' + message)

	return render(request, "login.html", context)

class NotificationCount(object):
	title = ""
	count = 0
	style = ""
	type = ''

	def __init__(self, title ="Unknown", count = 0, style = "fa fa-users text-aqua", type = 'web.views.appointments'):
		self.title = title
		self.count = count
		self.style = style
		self.type = type

def get_filter_locations(user):
	worker = None
	if isinstance(user, User):
		workers = Worker.objects.filter(user=user)
		if len(workers)>0:
			worker = workers[0]
	else:
		worker = user

	if worker==None or worker.locations.count() == 0:
		return Location.objects.all()
	else:
		return worker.locations.all()

def get_exceeded_visits(user):
	return Visit.objects.filter(datetime_end__lt = get_today_midnight_date(),
								is_finished = False,
								subject__default_location__in = get_filter_locations(user)
								).order_by('datetime_begin')

def get_exceeded_visit_notifications_count(user):
	notification = NotificationCount(
		title = "exceeded visit time",
		count = get_exceeded_visits(user).count(),
		style = "fa fa-thermometer-4 text-red",
		type = 'web.views.exceeded_visits')
	return notification


def get_subjects_with_no_visit(user):
	result = Subject.objects.annotate(my_count=Count(Case(When(visit__is_finished=False, then=1)))).filter(
		dead=False,
		resigned=False,
		my_count=0,
		default_location__in = get_filter_locations(user)
		)
	return result

def get_subject_with_no_visit_notifications_count(user):
	notification = NotificationCount(
		title = "subject without visit",
		count = get_subjects_with_no_visit(user).count(),
	 	style = "fa fa-users text-aqua",
		type = 'web.views.subject_no_visits')
	return notification


def get_visits_without_appointments_count(user):
	notification = NotificationCount(
		title = "unfinished visits ",
		count = get_unfinished_visits(user).count(),
		style = "fa fa-user-times text-yellow",
		type = 'web.views.unfinished_visits')
	return notification

def get_unfinished_visits(user):
	today = get_today_midnight_date()
	#list visits that have no scheduled appointments and taking place now
	return Visit.objects.annotate(my_count=Count(Case(When(appointment__status=Appointment.APPOINTMENT_STATUS_SCHEDULED, then=1)))).filter(
		datetime_begin__lt = today,
		datetime_end__gt = today,
		is_finished = False,
		subject__default_location__in = get_filter_locations(user),
		my_count=0)

def get_approaching_visits_without_appointments_count(user):
	notification = NotificationCount(
		title = "approaching visits ",
		count = get_approaching_visits_without_appointments(user).count(),
		style = "fa fa-users text-aqua",
		type = 'web.views.approaching_visits_without_appointments')
	return notification

def get_approaching_visits_without_appointments(user):
	today = get_today_midnight_date()
	today_plus_two_months =today+datetime.timedelta(days=62)
	return Visit.objects.annotate(my_count=Count(Case(When(appointment__status=Appointment.APPOINTMENT_STATUS_SCHEDULED, then=1)))).filter(
		datetime_begin__gt = today,
		datetime_begin__lt = today_plus_two_months,
		is_finished = False,
		subject__default_location__in = get_filter_locations(user),
		my_count=0)


def get_unfinished_appointments_count(user):
	return NotificationCount(
		title = "unfinished appointments ",
		count = get_unfinished_appointments(user).count(),
		style = "fa fa-history text-yellow",
		type = 'web.views.unfinished_appointments')

def get_unfinished_appointments(user):
	return Appointment.objects.filter(
		datetime_when__lt = get_today_midnight_date(),
		status = Appointment.APPOINTMENT_STATUS_SCHEDULED,
		location__in = get_filter_locations(user),
		)

def get_notifications(the_user):
	workers = Worker.objects.filter(user=the_user)
	notifications = []
	count = 0;
	if len(workers)>0:
		person = workers[0]
		if person.role == Worker.ROLE_CHOICES_SECRETARY:
			notifications.append(get_exceeded_visit_notifications_count(person))
			notifications.append(get_visits_without_appointments_count(person))
			notifications.append(get_approaching_visits_without_appointments_count(person))
			notifications.append(get_unfinished_appointments_count(person))
			notifications.append(get_subject_with_no_visit_notifications_count(person))
			for notification in notifications:
				count += notification.count
	return (count, notifications)

"""
Saturates response with information about logged user
"""
@login_required
def wrap_response(request, template, params):
	person, role = Worker.get_details(request.user)

	notifications = get_notifications(request.user)

	final_params = params.copy()
	final_params.update({
		'person' : person,
		'role': role,
		'notifications': notifications
	})

	return render(request, template, final_params)


def logout(request):
	state, message = do_logout(request)
	return redirect('/login?error=' + message)



def visits(request):
	visit_list = Visit.objects.order_by('-datetime_begin')
	context = {
		'visit_list': visit_list
	}

	return wrap_response(request, 'visits/index.html', context)

def exceeded_visits(request):
	context = {
		'visit_list': get_exceeded_visits(request.user)
	}
	return wrap_response(request, 'visits/index.html', context)

def unfinished_visits(request):
	context = {
		'visit_list': get_unfinished_visits(request.user)
	}

	return wrap_response(request, 'visits/index.html', context)

def approaching_visits_without_appointments(request):
	context = {
		'visit_list': get_approaching_visits_without_appointments(request.user)
	}

	return wrap_response(request, 'visits/index.html', context)

def visit_details(request, id):
	displayedVisit = get_object_or_404(Visit, id=id)
	visFinished = displayedVisit.is_finished
	vid = displayedVisit.id
	displayedSubject = displayedVisit.subject
	listOfAppointments = displayedVisit.appointment_set.all()
	canFinish=True
	for appointment in listOfAppointments:
		if appointment.status == Appointment.APPOINTMENT_STATUS_SCHEDULED:
			canFinish=False;
	vform = VisitDetailForm(instance=displayedVisit)
	sform = SubjectDetailForm(instance=displayedSubject)

	return wrap_response(request, 'visits/details.html', {
		'vform': vform,
		'sform': sform,
		'loApp': listOfAppointments,
		'visFinished': visFinished,
		'canFinish': canFinish,
		'vid': vid,
		'visit': displayedVisit})

def visit_mark(request, id, as_what):
	visit = get_object_or_404(Visit, id=id)
	if as_what == 'finished':
		visit.mark_as_finished()

	return redirect(visit_details, id=id)



def visit_add(request, subject_id = -1):
	if request.method == 'POST':
		form = VisitAddForm(request.POST, request.FILES)
		if form.is_valid():
			visit = form.save()
			return redirect(visit_details, visit.id)
	else:
		subjects = Subject.objects.filter(id=subject_id)
		subject = None
		if len(subjects)>0:
			subject = subjects[0]
		form = VisitAddForm(initial={'subject':subject})


	return wrap_response(request, 'visits/add.html', {'form': form})


def subjects(request):
	subjects_list = Subject.objects.order_by('-last_name')
	context = {
		'subjects_list': subjects_list
	}

	return wrap_response(request, 'subjects/index.html', context)

def subject_add(request):
	if request.method == 'POST':
		form = SubjectAddForm(request.POST, request.FILES)
		if form.is_valid():
			screening_number = form.cleaned_data['screening_number']
			if screening_number == '':
				screening_number = get_new_screening_number()
			form.save()
			return redirect(subjects)
	else:
		form = SubjectAddForm()

	return wrap_response(request, 'subjects/add.html', {'form': form})

def subject_no_visits(request):
	subjects_list = get_subjects_with_no_visit(request.user).order_by('-last_name')
	context = {
		'subjects_list':  subjects_list
	}

	return wrap_response(request, 'subjects/index.html', context)

def subject_edit(request, id):
	the_subject = get_object_or_404(Subject, id=id)
	if request.method == 'POST':
		form = SubjectEditForm(request.POST, request.FILES, instance=the_subject)
		if form.is_valid():
			form.save()
			return redirect(subjects)
	else:
		form = SubjectEditForm(instance=the_subject)
	return wrap_response(request, 'subjects/edit.html', {
		'form': form,
		'subject': the_subject
		})


def subject_delete(request, id):
    the_subject = get_object_or_404(Subject, id=id)
    if request.method == 'POST':
        the_subject.delete()
        return redirect(subjects)
    else:
        form = SubjectEditForm(instance=the_subject)
    return wrap_response(request, 'subjects/delete.html', {'form': form})


def subject_mark(request, id, as_what):
	who = get_object_or_404(Subject, id=id)
	if as_what == 'dead':
		who.mark_as_dead()
	elif as_what == 'rejected':
		who.mark_as_rejected()
	return redirect(subject_edit, id=id)

def appointment_mark(request, id, as_what):
	appointment = get_object_or_404(Appointment, id=id)
	if as_what == 'finished':
		appointment.mark_as_finished()
	elif as_what == 'cancelled':
		appointment.mark_as_cancelled()
	elif as_what == 'no_show':
		appointment.mark_as_no_show()
	else:
		return e500_error(request)
	return redirect(visit_details, id=appointment.visit.id)

def subject_visit_details(request, id):
	locsubject = get_object_or_404(Subject, id=id)
	visits = locsubject.visit_set.all()
	endlist = []
	for vis in visits:
		assign = vis.appointment_set.all()
		finished = vis.is_finished
		visid = vis.id
		visit_title = vis.follow_up_title()
		visform = VisitDetailForm(instance=vis)
		endlist.append((visform,assign,finished,visid,visit_title))

	return wrap_response(request, 'subjects/visitdetails.html', {'display': endlist, "id":id})

def doctors(request):
	doctors_list = Worker.objects.order_by('-last_name')
	context = {
		'doctors_list': doctors_list
	}

	return wrap_response(request, "doctors/index.html", context)


def doctor_add(request):
	if request.method == 'POST':
		form = WorkerAddForm(request.POST, request.FILES)
		if form.is_valid():
			form.save()
			return redirect(doctors)
	else:
		form = WorkerAddForm()

	return wrap_response(request, 'doctors/add.html', {'form': form})



def doctor_edit(request, doctor_id):
	the_doctor = get_object_or_404(Worker, id=doctor_id)
	if request.method == 'POST':
		form = WorkerEditForm(request.POST, request.FILES, instance=the_doctor)
		if form.is_valid():
			form.save()
			return redirect(doctors)
	else:
		form = WorkerEditForm(instance=the_doctor)
	return wrap_response(request, 'doctors/edit.html', {'form': form})


def doctor_details(request, doctor_id):
	the_doctor = get_object_or_404(Worker, id=doctor_id)
	form = WorkerDetailForm(instance=the_doctor)

	return wrap_response(request, 'doctors/details.html', {'form': form})


def doctor_availability(request, doctor_id):
	avall = Avaibility.objects.filter(person=doctor_id)

	avmon = avall.filter(day_number=1)
	avtue = avall.filter(day_number=2)
	avwed = avall.filter(day_number=3)
	avthu = avall.filter(day_number=4)
	avfri = avall.filter(day_number=5)
	avsat = avall.filter(day_number=6)
	avsun = avall.filter(day_number=7)

	context = {
		'avmon': avmon,
		'avtue': avtue,
		'avwed': avwed,
		'avthu': avthu,
		'avfri': avfri,
		'avsat': avsat,
		'avsun': avsun,
		'id': doctor_id
	}

	return wrap_response(request, "doctors/availability_index.html", context)


def doctor_availability_delete(request, doctor_id, availability_id):
	availibility = Avaibility.objects.filter(id=availability_id)
	if len(availibility) > 0:
		availibility.delete()
	return redirect(doctoravail, id=doctor_id)


def equipment_def(request):
	equipment_list = Item.objects.order_by('-name')
	context = {
		'equipment_list': equipment_list
	}

	return wrap_response(request, "eqdef/index.html", context)


def equipment_and_rooms(request):
	return wrap_response(request, "equipment_and_rooms/index.html", {})


def mail_templates(request):
	return wrap_response(request, "mail_templates/index.html", {})


"""
#An initial draft of a function that was supposed to suggest date, room and worker for an appointment
def suggest_details(Appointment appoint):

	avaibleWorkers = Worker.objects.get()
	if appoint.appointment_type__required_worker == 'DOCTOR':
		avaibleWorkers.filter(role='DOCTOR')
	elif appoint.appointment_type__required_worker == 'NURSE':
		avaibleWorkers.filter(role__in=['DOCTOR','NURSE'])
	elif appoint.appointment_type__required_worker == 'PSYCHOLOGIST':
		avaibleWorkers.filter(role__in=['DOCTOR','PSYCHOLOGIST'])

	avaibleRooms = Room.objects.get
	requireditems = appoint.appointment_type.required_equipment.filter(is_fixed=True)
	reduce(operator.and_, (Q(equipment__contains=requireditems) for x in avaibleRooms))

"""

def get_today_midnight_date():
	today = datetime.datetime.now()
	today_midnight = datetime.datetime(today.year,today.month,today.day)
	return today_midnight

def get_calendar_full_appointments(user):
	month_ago = get_today_midnight_date() + datetime.timedelta(days=-31)
	return Appointment.objects.filter(
			datetime_when__gt = month_ago,
			location__in = get_filter_locations(user),
			).order_by('datetime_when')


def appointments(request):
	approaching_list = Appointment.objects.filter(
		datetime_when__gt = get_today_midnight_date(),
		location__in = get_filter_locations(request.user),
		status = Appointment.APPOINTMENT_STATUS_SCHEDULED
		).order_by('datetime_when')

	full_list = get_calendar_full_appointments(request.user)

	context = {
		'approaching_list': approaching_list,
		'full_list': full_list
	}

	return wrap_response(request, "appointments/index.html",context)

def unfinished_appointments(request):
	appointments = get_unfinished_appointments(request.user)
	context = {
		'appointment_list': appointments,
	}

	return wrap_response(request, "appointments/list.html",context)

def appointment_details(request, id):
	the_appointment = get_object_or_404(Appointment, id=id)
	form = AppointmentDetailForm(instance=the_appointment)
	return wrap_response(request, 'appointments/details.html', {'form': form})


def appointment_add(request, id):
	full_list = get_calendar_full_appointments(request.user)
	if request.method == 'POST':
		form = AppointmentAddForm(request.POST, request.FILES)
		form.fields['visit'].widget = forms.HiddenInput()
		if form.is_valid():
			form.save()
			return redirect(visit_details, id=id)
	else:
		form = AppointmentAddForm(initial={'visit': id})
		form.fields['visit'].widget = forms.HiddenInput()

	return wrap_response(request, 'appointments/add.html', {'form': form, 'visitID': id, 'full_appointment_list': full_list})

def appointment_edit(request, id):
	the_appointment = get_object_or_404(Appointment, id=id)
	if request.method == 'POST':
		form = AppointmentEditForm(request.POST, request.FILES, instance=the_appointment)
		if form.is_valid():
			form.save()

			data = form.cleaned_data
			vis = data['visit']
			visit = get_object_or_404(Visit, id=vis.id)

			return redirect(appointments)
	else:
		form = AppointmentEditForm(instance=the_appointment)

	subject_form = SubjectDetailForm(instance=the_appointment.visit.subject)

	return wrap_response(request, 'appointments/edit.html', {
		'form': form,
		'subject_form': subject_form,
		'id':id,
		'appointment': the_appointment
		})


def appointment_edit_datetime(request, id):
	the_appointment = get_object_or_404(Appointment, id=id)
	if request.method == 'POST':
		form = AppointmentEditForm(request.POST, request.FILES, instance=the_appointment)
		if form.is_valid():
			form.save()
			return redirect(appointments)
	else:
		the_appointment.datetime_when = the_appointment.visit.datetime_begin
		form = AppointmentEditForm(instance=the_appointment)
	return wrap_response(request, 'appointments/edit.html', {'form': form})

def export_to_csv2(request, type="subjects"):
    #Create the HttpResponse object with the appropriate CSV header.
	response = HttpResponse(content_type='text/csv')
	response['Content-Disposition'] = 'attachment; filename="'+type+'-'+get_today_midnight_date().strftime("%Y-%m-%d")+'.csv"'

	writer = csv.writer(response, quotechar =str(u'"'), quoting = csv.QUOTE_ALL)
	if type=="subjects":
		write_subjects_to_csv(writer)
	elif type=="appointments":
		write_appointments_to_csv(writer)
	else:
		return e500_error(request)
	return response

def write_subjects_to_csv(writer):
	subject_fields = []
	for field in Subject._meta.fields:
		if  field.name!="ID":
			subject_fields.append(field)

	field_names = []
	for field in subject_fields:
		field_names.append(field.verbose_name)

	writer.writerow(field_names)

	subjects = Subject.objects.order_by('-last_name')
	for subject in subjects:
		row = []
		for field in subject_fields:
			row.append(getattr(subject,field.name))
		writer.writerow([unicode(s).encode("utf-8") for s in row])

def write_appointments_to_csv(writer):
	appointments_fields = []
	for field in Appointment._meta.fields:
		if field.name!="visit" and field.name!="id" and	field.name!="worker_assigned" and field.name!="appointment_types" and field.name!="room" and field.name!="flying_team":
			appointments_fields.append(field)

	field_names = []
	field_names.append('ND number')
	field_names.append('Family name')
	field_names.append('Name')
	field_names.append('Visit')
	for field in appointments_fields:
		field_names.append(field.verbose_name)

	writer.writerow(field_names)

	appointments = Appointment.objects.order_by('-datetime_when')

	for appointment in appointments:
		row = []
		row.append(appointment.visit.subject.nd_number)
		row.append(appointment.visit.subject.last_name)
		row.append(appointment.visit.subject.first_name)
		row.append(appointment.visit.follow_up_title())
		for field in appointments_fields:
			row.append(getattr(appointment,field.name))
		type_string = ""
		for type in appointment.appointment_types.all():
			type_string+=type.code+","
		row.append(type_string)
		writer.writerow([unicode(s).encode("utf-8") for s in row])

def export(request):
	return wrap_response(request, 'export/index.html',{})
